from unittest import TestCase
from pypass.generator.diceware import Diceware

import collections
import sqlite3
import re


class TestDiceware(TestCase):
    def test_create(self):
        obj = Diceware()
        self.assertIsInstance(obj, Diceware, "Did not create a Diceware object")
        self.assertIsInstance(obj, collections.Mapping, "Diceware object not inherited from collections.Mapping")
        self.assertIsInstance(obj._database, sqlite3.Connection, "Diceware database is not a SQLite connection")

    def test_lookup(self):
        dice = Diceware()
        self.assertTrue(getattr(dice, 'word'))
        self.assertEqual('abash', dice.word(11131), "Unable to lookup word by int")
        self.assertEqual('abash', dice.word('11131'), "Unable to lookup word by str")

        self.assertEqual('abash', dice[11131], "Unable to dict-style lookup by int")
        self.assertEqual('abash', dice['11131'], "Unable to dict-style lookup by str")

    def test_dict_performance(self):
        dice = Diceware()

        # lookups
        self.assertEqual('abash', dice[11131], "Unable to dict-style lookup by int")
        self.assertEqual('abash', dice['11131'], "Unable to dict-style lookup by str")

        # len
        self.assertEqual(7776, len(dice), "Len did not return 7776")

        # iter
        i = iter(dice)
        self.assertEqual(11111, i.next(), "First iter value was not all 1")
        for x in range(1, 6):
            i.next()

        self.assertEqual(11121, i.next(), "7th value was not 11121")

    def test_roll(self):
        dice = Diceware()
        self.assertTrue(re.search("^[1-6]{5}$", dice.roll()))

    def test_random_word(self):
        dice = Diceware()
        self.assertTrue(getattr(dice, 'random'))
        self.assertTrue(re.search('^\S+$', dice.random()), "Random selection doesn't appear to be a single word")
        self.assertTrue(type(dice.random(3)) is list, "Random selection of multiple words does not return a list")
        self.assertRaises(ValueError, dice.random, -1)



"""
pypass.generator.diceware - contains a diceware wordlist interface along with a generator

NOTE: the generator somewhat defeats the purpose of using diceware, use with caution
"""
from __future__ import print_function

import collections
import re
import sqlite3
from os import path
from random import SystemRandom

import requests


class Diceware(collections.Mapping):
    """
    A representation of the Diceware word list

    Uses a SQLite database under the covers to avoid having a whole structure in memory. Access options:

    dice = Diceware()
    dice.word(id) - provide a 5-die roll as int (e.g. 11346) or str (e.g. "11346"), returns that word
    dice[id] - another way to access .word(id)

    dice.random() - get a random word as a string
    dice.random(5) - get several random words as a list of strings

    Random number generation is done with a class-level instance of random.SystemRandom(), which uses urandom
    under the covers for cryptographically-sufficent randomness
    """
    valid_id_re = re.compile("^[1-6]{5}$")
    rng = SystemRandom()

    def __init__(self):
        self._dbfile = path.join(path.dirname(__file__), 'diceware.sqlite')
        self._database = sqlite3.connect(self._dbfile)

    def word(self, index):
        """
        Returns a word corresponding to a given diceware roll (a sequence of five d-6 rolls)

        :param index: a str or int representing the roll, e.g. 16254
        :return str: the word corresponding to the roll
        """
        with self._database as conn:
            cur = conn.cursor()
            cur.execute("SELECT word FROM words WHERE id = ?", (index,))
            results = cur.fetchall()

        if len(results) <= 0:
            return None
        elif len(results) > 1:
            raise RuntimeError("Expected exactly 1 result, got {}; database may be corrupt".format(len(results)))
        else:
            return results[0][0]

    def _roll_die(self):
        """
        Rolls a single die
        :return int: between 1 and 6, inclusive of both
        """
        return self.rng.randint(1, 6)

    def roll(self):
        """
        Produce a diceware roll for one word. Rolls a die 5 times

        :return str: a string representing the roll, e.g. "45632"; always length=5
        """
        rollstr = ""
        for a in range(0, 5):
            rollstr += str(self._roll_die())
        return rollstr

    def random(self, count=1):
        """
        Produces one or more random diceware words from the list.

        For each word to be selected, uses .roll() to generate a roll.

        :param count: and integer that specifies the number of random words to choose
        :return str: if 1 word is requested (default), returns that word as a string
        :return list: if more than one word is requested, returns a list of strings
        """
        if count == 1:
            return self.word(self.roll())
        elif count > 1:
            words = []
            for x in range(0, count):
                words.append(self.random(1))
            return words
        raise ValueError("Must ask for a positive number of words")

    def __getitem__(self, key):
        return self.word(self.__keytransform__(key))

    def __iter__(self):
        for a in range(1, 7):
            for b in range(1, 7):
                for c in range(1, 7):
                    for d in range(1, 7):
                        for e in range(1, 7):
                            yield int("{}{}{}{}{}".format(a, b, c, d, e))

    def __len__(self):
        with self._database as conn:
            cur = conn.cursor()
            cur.execute("SELECT count(id) FROM words")
            result = cur.fetchone()
            return result[0]

    def __keytransform__(self, key):
        if not self.valid_id_re.search(str(key)):
            raise KeyError("'{}' is not a valid diceware index")
        return int(key)


# noinspection PyProtectedMember
def _create_schema(dw_obj=None):
    """
    Utility function to create the SQLite database schema. Shouldn't need to use unless you're creating
    your own, non-default database for some reason (such as from an alternate word list)

    :param dw_obj: a Diceware object. If None, instantiates one with defaults
    :return:
    """
    if dw_obj is None:
        dw_obj = Diceware()

    with dw_obj._database as conn:
        conn.execute("CREATE TABLE WORDS (id INTEGER PRIMARY KEY, word TEXT NOT NULL)")
        conn.commit()


# noinspection PyProtectedMember
def _fill_database(dw_obj=None, wordlist_url='http://world.std.com/~reinhold/diceware.wordlist.asc'):
    """
    Downloads the wordlist file and inserts the words into an existing SQLite database.

    By default, fetches the official diceware wordlist; you can alter this, by pointing to a URL with a
    diceware-format alternate text file.

    :param dw_obj: a Diceware object. If None, instantiates one with defaults
    :param wordlist_url: a URL pointing at a text file containing space-separated rows of roll and word
    :return:
    """
    if dw_obj is None:
        dw_obj = Diceware()

    response = requests.get(wordlist_url)

    with dw_obj._database as conn:
        line_ok_regex = re.compile("^(\d{5})\s+(.+)\s*$")
        line_count = 0
        for line in response.text.splitlines():
            match = line_ok_regex.search(line)
            if not match:
                continue  # line does not contain a diceware word
            index = match.group(1).strip()
            word = match.group(2).strip().lower()
            conn.execute("INSERT INTO words (id, word) VALUES (?, ?)", (int(index), word))
            line_count += 1
            if line_count % 50 == 0:
                print("X", end="")
                conn.commit()

        print("")
        conn.commit()



from unittest import TestCase

from pypass.generator import Diceware
from pypass.passphrase import PassphraseGenerator


class MockGenerator(object):
    def __init__(self):
        pass

    def random(self, count=1):
        words = "the quick brown fox jumped over the lazy dog".split(' ')

        if count < 1:
            raise ValueError("Requires a positive count")
        if count == 1:
            return words[0]
        else:
            retvalue = words[0:count]
            return retvalue


class TestPassphraseGenerator(TestCase):
    def testCreate(self):
        generator = PassphraseGenerator()
        self.assertIsInstance(generator, PassphraseGenerator)
        self.assertIsInstance(generator.generator, Diceware, "Doesn't default to Diceware generator")
        self.assertEqual(5, generator.elements, "Does not default to 5-word phrase")

    def testCreateSetLength(self):
        generator = PassphraseGenerator(elements=3)
        self.assertEqual(3, generator.elements, "Does not set number of elements with constructor arg")

    def testCreateSetGenerator(self):
        generator = PassphraseGenerator(generator=MockGenerator())
        self.assertIsInstance(generator.generator, MockGenerator, "Doesn't allow custom generator at construction")

    def testGeneration(self):
        generator = PassphraseGenerator(generator=MockGenerator(), elements=3)
        phrase = generator.generate()
        self.assertEqual("thequickbrown", phrase, "Doesn't generate expected phrase, got '{}'".format(phrase))

    def testGenerateWithCustomSeparator(self):
        generator = PassphraseGenerator(generator=MockGenerator(), elements=3, separator='-')
        phrase = generator.generate()
        self.assertEqual("the-quick-brown", phrase, "Custom sep in constructor failed, got '{}'".format(phrase))

        phrase = generator.generate(separator='.')
        self.assertEqual("the.quick.brown", phrase, "Custom sep in method failed, got '{}'".format(phrase))

    def testGenerateWithCustomElements(self):
        generator = PassphraseGenerator(generator=MockGenerator(), elements=4)
        phrase = generator.generate()
        self.assertEqual("thequickbrownfox", phrase, "Custom count in constructor failed, got '{}'".format(phrase))

        phrase = generator.generate(elements=3)
        self.assertEqual("thequickbrown", phrase, "Custom count in method failed, got '{}'".format(phrase))

    def testCustomWordTransform(self):
        def capitalize_first(word):
            return word.capitalize()

        def all_caps(word):
            return word.upper()

        generator = PassphraseGenerator(generator=MockGenerator(), elements=2, word_transform=capitalize_first)
        phrase = generator.generate()
        self.assertEqual("TheQuick", phrase, "Custom word transform in constructor failed, got '{}'".format(phrase))

        phrase = generator.generate(word_transform=all_caps)
        self.assertEqual("THEQUICK", phrase, "Custom word transform in method failed, got '{}'".format(phrase))

    def testCustomPhraseTransform(self):
        def add_numbers(phrase):
            return phrase + "42"

        def all_caps(phrase):
            return phrase.upper()

        generator = PassphraseGenerator(generator=MockGenerator(), elements=3, phrase_transform=all_caps)
        phrase = generator.generate()
        self.assertEqual("THEQUICKBROWN", phrase,
                         "Custom phrase transform in constructor failed, got '{}'".format(phrase))

        phrase = generator.generate(phrase_transform=add_numbers)
        self.assertEqual("thequickbrown42", phrase,
                         "Custom phrase transform in method failed, got '{}'".format(phrase))











"""
Passphrase generator for python

Uses various backends in the pypass.generator.* namespace. A generator object should have a
.random() method that:

 - returns a single, randomly-chosen word
 - returns a list of words if passed a number > 1 as a parameter
"""

from pypass.generator import diceware


class PassphraseGenerator(object):
    """
    Generates passphrases, by default using Diceware

    Allows you to specify a generator (see Generators, below), the number of words, the separator
    to use between them, and callables to transform each word and/or each phrase (see ``.transform.*``)

    By default, generates a 5-word diceware phrase with no separators, and uses no transforms

    Word count (``elements``), separator, word transforms, and phrase transforms can all be specified at
    construction time, or be modified for an instance by changing the object's attributes of the same names.
    Each can also be overridden for a specific call to ``generate()`` by passing arguments to that method.

    Generators:

    A generator is an object that has a `random` method taking an argument for the number of words to
    generate. The `random` method should return a single word by default or if the number of words is 1,
    otherwise it should return a list of words of the requested length.

    See ``pypass.generator.*``

    Transforms:

    A transform is a callable that accepts a string as an argument and returns a transformed version of it.
    See the transform.* functions for generators of common transforms.

    The ``word_transform`` will be called on each word generated as it's added to the phrase. Separators
    are not passed to word transforms.

    The ``phrase_transform`` will be called on the generated phrase before returning it.

    Examples:

        from pypass.passphrase import PassphraseGenerator, transform

        generator = PassphraseGenerator(elements=7, separator='-')
        passphrase = generator.generate()  # 7-word passphrase, separated by '-'

        generator.word_transform = transform.uppercase()  # next generation, each word will be uppercased
    """
    def __init__(self, elements=5, separator='', **kwargs):
        """
        Constructor - configures the object (by default: 5-word, no separator, Diceware)

        Keywords:

        - generator: specify a generator object to use instead of the default
        - word_transform: a callable that will transform each generated word
        - phrase_transform: a callable that will transform the entire phrase

        :param elements: number of words to generate
        :param separator: separator to use
        :param kwargs: configuration arguments
        """
        self.generator = None
        self.elements = elements
        self.separator = separator
        self.word_transform = None
        self.phrase_transform = None

        if 'generator' in kwargs:
            self.generator = kwargs['generator']

        if 'word_transform' in kwargs:
            self.word_transform = kwargs['word_transform']

        if 'phrase_transform' in kwargs:
            self.phrase_transform = kwargs['phrase_transform']

        if self.generator is None:
            self.generator = diceware.Diceware()

    def generate(self, elements=None, separator=None, word_transform=None, phrase_transform=None):
        """
        Generates a passphrase string.

        With no arguments provided, will use the object's settings; these can be overriden on a given
        call using parameters.

        :param elements: number of words for the generator to use
        :param separator: character to place between words
        :param word_transform: transform callable for each word
        :param phrase_transform: transform callable for the entire phrase
        :return: passphrase string
        """
        if elements is None:
            elements = self.elements

        if separator is None:
            separator = self.separator

        if word_transform is None:
            word_transform = self.word_transform

        if phrase_transform is None:
            phrase_transform = self.phrase_transform

        words = self.generator.random(elements)
        if type(words) is not list:
            words = [words]

        if self.word_transform is not None:
            for i in range(0, len(words)):
                words[i] = word_transform(words[i])

        phrase = separator.join(words)

        if self.phrase_transform is not None:
            phrase = phrase_transform(phrase)

        return phrase

    # TODO strength? May need to modify Generator contract



import re
from unittest import TestCase

from pypass.passphrase import transform


class TestTransform(TestCase):
    def testAddNumbers(self):
        adder = transform.add_numbers(3)
        teststring = adder("hello")
        self.assertTrue(re.search("^hello\d\d\d$", teststring),
                        "Did not add three numbers, got '{}'".format(teststring))

    def testCapitalize(self):
        capper = transform.capitalize()
        testtring = capper("hello")
        self.assertEqual("Hello", testtring, "Didn't capitalize word, got '{}'".format(testtring))

    def testLeet(self):
        leeter = transform.leet(probability=1)
        teststring = leeter("hello")
        self.assertEqual("h3110", teststring, "Didn't leetify word, got '{}'".format(teststring))

        teststring = leeter("aeiout")
        self.assertEqual("43!0u+", teststring, "Didn't leetify word, got '{}'".format(teststring))

    def testChain(self):
        chainer = transform.chain(transform.capitalize(), transform.add_numbers(2))
        teststring = chainer("hello")
        self.assertTrue(re.search("^Hello\d\d$", teststring), "Didn't do both functions, got '{}'".format(teststring))



from setuptools import setup
from os import path

generator_root = path.join('pypass', 'generator')

setup(
    name='pypass-generator',
    version='0.1',
    packages=['pypass', 'pypass.generator', 'pypass.passphrase'],
    # data_files=[(generator_root, [path.join(generator_root, 'diceware.sqlite')])],
    package_data = {'pypass.generator': ['diceware.sqlite']},
    include_package_data = True,
    install_requires=['requests>=2.12.5'],
    url='https://bitbucket.org/darrenpmeyer/pypass-generator',
    license='License :: OSI Approved :: Apache Software License',
    author='Darren P Meyer',
    author_email='darren@darrenpmeyer.com',
    description='Password and Passphrase Generator using Diceware and other sources'
)



"""
Provides transformation generators for use with PassphraseGenerator's word_transform and phrase_transform
callables

Examples:

    from pypass.passphrase import transform

    # Add 3 random digits to each generated word
    generator = PassphraseGenerator(word_transform=transform.add_numbers(3))

    # Capitalize the first letter of each word
    generator = PassphraseGenerator(word_transform=transform.capitalize())
"""
from random import SystemRandom


def add_numbers(count):
    """
    Generates a word/phrase transform that adds ``count`` numbers to the end of a string

    :param count: how many numbers to add
    :return: a function suitable for use as a transform
    """

    rng = SystemRandom()

    def add_number_transform(word):
        retvalue = word
        for n in range(0, count):
            retvalue += str(rng.randint(0, 9))
        return retvalue

    return add_number_transform


def capitalize():
    """
    Generates a word/phrase transform that capitalizes the first letter of a string

    It might seem weird for this to be a generator, but it keeps it consistent with the interfaces
    in this module

    :return:
    """
    def capify(item):
        return item.capitalize()

    return capify


def leet(probability=0.2):
    """
    Generates a word/phrase transform that "l33tifies" a string.

    :param probability: float - probability of a subsitutable letter getting changed (default=0.2 [20%])
    :return:
    """
    rng = SystemRandom()

    # TODO more substititions
    substitutions = {
        'a': '4',
        'e': '3',
        'i': '!',
        'o': '0',
        'l': '1',
        't': '+'
    }

    def leetify(item):
        retval = ''
        for letter in item:
            nl = letter
            if letter in substitutions and rng.random() < probability:
                nl = substitutions[letter]
            retval += nl

        return retval

    return leetify


def chain(*args):
    """
    Generates a transform that consists of a chain of other transforms

    Example:

        my_transform = chain(add_numbers(2), capitalize())
        generator.generate(word_transform=my_transform)  # e.g. "Hello12There27"

    :param args: other transforms
    :return: callable
    """

    if len(args) < 1:
        raise ValueError("Need at least one transform to chain")

    def chainer(item):
        retval = item
        for transform in args:
            retval = transform(retval)

        return retval

    return chainer
